/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include "BlackScholes.h"
#include <sys/mman.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <fenv.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/time.h>  // gettimeofday
#include <math.h>
#include <string.h>
#include "alf.h"

#define A1  0.319381530
#define A2 -0.356563782
#define A3  1.781477937
#define A4 -1.821255978
#define A5  1.330274429
#define INV_ROOT2PI 0.39894228

/* next two lines used for huge page size stuff */
char *mem_file = "/huge/BlackScholes_mem.bin";
char *mem_addr = NULL;

int NP, NP_BYTES, N_SPUS, LN2_N_SPUS;
int totmem, chunksize_bytes, chunksize_samples, bytes_per_sample, double_flag;

void *cps[MAX_SPUS];
void *S0[MAX_SPUS], *K[MAX_SPUS], *r[MAX_SPUS], *sigma[MAX_SPUS], *T[MAX_SPUS], *answer[MAX_SPUS];

struct timeval timev1, timev2; // gettimeofday

#define PATH_BUF_SIZE 1024                // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE + 1];   // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE + 1];     // Used to hold the name of spu library
char spu_image_name[] = "BlackScholes_spu";
char kernel_name[] = "comp_kernel";
char input_dtl_name[] = "input_list_prepare";
char output_dtl_name[] = "output_list_prepare";

#define MALLOC_BIG_ARRAYS {                                 \
  unsigned long long raw;                                   \
  PTR_TO_ADDR64(malloc(totmem + 0x80), raw);                \
  if (raw == 0) {                                           \
    printf("ERROR: unable to malloc.  Exiting...\n");       \
    return(-1);                                             \
  }                                                         \
  raw += 0x7f;                                              \
  raw &= (~0x7f);                                           \
  ADDR64_TO_PTR(raw, char *, mem_addr);                     \
}

double N(double x) {
   double k, n;
   double accum;
   double candidate_answer;
   int flag;
   flag = (x < 0);
   x = (x < 0) ? -x : x;
   k = 1.0 / (1.0 + 0.2316419 * x);
   accum = A4 + A5 * k;
   accum = k * accum + A3;
   accum = k * accum + A2;
   accum = k * accum + A1;
   accum = k * accum;
   n = exp(-0.5 * x * x);
   n *= INV_ROOT2PI;
   candidate_answer = 1.0 - n * accum;
   return (flag ? 1.0 - candidate_answer : candidate_answer);
}

double bspf(unsigned long long cpflag, double S0, double K, double r, double sigma, double T) {
   double d1, d2, c, p, Nd1, Nd2, expval, answer;
   d1 = log(S0/K) + (r + 0.5*sigma*sigma)*T;
   d1 /= (sigma * sqrt(T));
   expval = exp(-r * T);
   d2 = d1 - sigma * sqrt(T);
   Nd1 = N(d1);
   Nd2 = N(d2);
   c = S0 * Nd1 - K * expval * Nd2;
   p = K * expval * (1.0 - Nd2) - S0 * (1.0 - Nd1);
   answer = cpflag ? c : p;
   return answer;
}

int main(int argc, char *argv[]) {
  void *config_parms = NULL;
  int rc;
  
  alf_handle_t alf_handle;
  alf_task_handle_t task_handle;
  alf_task_desc_handle_t task_desc_handle;
  alf_wb_handle_t wb_handle;
#ifdef _ALF_PLATFORM_HYBRID_
  alf_dataset_handle_t dataset_handle;
#endif /* _ALF_PLATFORM_HYBRID_ */

  BlackScholesPara BSPara __attribute__  ((aligned(128)));
  
  int i, j, n;
  long long int n_cycles;
  int fmem;
  double lo_error, hi_error;
  float time_seconds;
  unsigned int count_spes;
  unsigned int nloops;
  unsigned int multi_use_wb_count;
  

  if (argc != 5) {
    fprintf(stderr, "usage: BlackScholes <LOG_N_SPUS> <double_flag> <ncycles> <size, in 64K chunks>\n");
    return -1;
  }
  LN2_N_SPUS = atoi(argv[1]);
  if (LN2_N_SPUS < 0 || LN2_N_SPUS > 4) {
     printf("LOG_N_SPUS must be between 0 and 4\n");
     return -1;
  }
  N_SPUS = 1 << LN2_N_SPUS;

  double_flag = atoi(argv[2]);
  if (double_flag < 0 || double_flag > 1) {
     printf("double_flag must be 0 or 1\n");
     return -1;
  }
  bytes_per_sample = 1 << (2+double_flag);

  n_cycles = atoi(argv[3]);
  if (n_cycles < 0) {
     printf("n_cycles can not be less than 0.\n");
     return -1;
  }

  NP = atoi(argv[4]);
  NP <<= 16;
  if (NP > MAX_NP_BYTES / bytes_per_sample) {
     printf("truncating size from %d to %d\n", NP, MAX_NP_BYTES / bytes_per_sample);
     NP = MAX_NP_BYTES / bytes_per_sample;
  }
  NP_BYTES = NP * bytes_per_sample;
  if (NP == 0) {
     printf("size must be nonzero\n");
     return -1;
  }

  totmem = (NP_BYTES) * 7;  /* 7 different arrays */

  printf ("number of bytes per buffer = %d\n", NP_BYTES);

  chunksize_bytes = (NP_BYTES / N_SPUS);
  chunksize_samples = (NP / N_SPUS);

  /* Create a large contiguous memory buffer by allocating a large
   * page (or more). Large page memory will also reduce the TLB thrashing.
   */
  if ((fmem = open (mem_file, O_CREAT | O_RDWR, 0755)) == -1) {
    printf("WARNING: unable to open file %s (errno=%d). Using malloc heap.\n", mem_file, errno);
    MALLOC_BIG_ARRAYS
  } else {
    int lp_totmem;
    lp_totmem = (totmem + 0xffffff) & (~0xffffff);
    mem_addr = (char *) mmap (0, lp_totmem, PROT_READ | PROT_WRITE, MAP_SHARED, fmem, 0);
    if (mem_addr == MAP_FAILED) {
      printf("ERROR: unable to mmap file %s (errno=%d). Using malloc heap.\n", mem_file, errno);
      close (fmem);
      MALLOC_BIG_ARRAYS
    }
  }

  for (i = 0; i < N_SPUS; ++i)
  {
    cps[i]      = (void*) (mem_addr + (i * chunksize_bytes));
    S0[i]       = (void*) (mem_addr + ((1 * N_SPUS) + i) * chunksize_bytes);
    K[i]        = (void*) (mem_addr + ((2 * N_SPUS) + i) * chunksize_bytes);
    r[i]        = (void*) (mem_addr + ((3 * N_SPUS) + i) * chunksize_bytes);
    sigma[i]    = (void*) (mem_addr + ((4 * N_SPUS) + i) * chunksize_bytes);
    T[i]        = (void*) (mem_addr + ((5 * N_SPUS) + i) * chunksize_bytes);
    answer[i]   = (void*) (mem_addr + ((6 * N_SPUS) + i) * chunksize_bytes);
  }


  printf("loading big arrays...\n"); fflush(stdout);

  {
    int S0Kdex, rdex, sigdex, Tdex;
    if (double_flag) {
      double S0_array[4] = {42.0, 30.0, 54.0, 66.0};
      double K_array[16] = {40.0, 36.0, 44.0, 48.0,
                            24.0, 28.0, 32.0, 36.0,
                            48.0, 52.0, 56.0, 60.0,
                            60.0, 64.0, 68.0, 72.0};
      double r_array[4] = {0.1, 0.09, 0.11, 0.12};
      double sigma_array[4] = {0.2, 0.15, 0.25, 0.30};
      double T_array[4] = {0.5, 0.25, 0.75, 1.0};
      double *Tptr, *sigptr, *rptr, *S0ptr, *Kptr, *answerptr;
      unsigned long long *cpsptr;
      for (j=0; j<N_SPUS; ++j) {
        Tptr = (double *) T[j];
        sigptr = (double *) sigma[j];
        S0ptr = (double *) S0[j];
        Kptr = (double *) K[j];
        answerptr = (double *) answer[j];
        rptr = (double *) r[j];
        cpsptr = (unsigned long long *) cps[j];
        for (i=0; i<chunksize_samples; ++i) {
          Tdex   = (i >> 1) & 0x3;
          sigdex = (i >> 3) & 0x3;
          rdex   = (i >> 5) & 0x3;
          S0Kdex = (i >> 7) & 0xf;
          *cpsptr++ = (i&1) ? 0xffffffffffffffffULL : 0ULL;
          *S0ptr++ = S0_array[S0Kdex >> 2];
          *Kptr++ = K_array[S0Kdex];
          *rptr++ = r_array[rdex];
          *sigptr++ = sigma_array[sigdex];
          *Tptr++ = T_array[Tdex];
          *answerptr++ = 0.0;
        }
      }
    }
    else {
      float S0_array[4] = {42.0, 30.0, 54.0, 66.0};
      float K_array[16] = {40.0, 36.0, 44.0, 48.0,
                           24.0, 28.0, 32.0, 36.0,
                           48.0, 52.0, 56.0, 60.0,
                           60.0, 64.0, 68.0, 72.0};
      float r_array[4] = {0.1, 0.09, 0.11, 0.12};
      float sigma_array[4] = {0.2, 0.15, 0.25, 0.30};
      float T_array[4] = {0.5, 0.25, 0.75, 1.0};
      float *Tptr, *sigptr, *rptr, *S0ptr, *Kptr, *answerptr;
      unsigned int *cpsptr;
      for (j=0; j<N_SPUS; ++j) {
        Tptr = (float *) T[j];
        sigptr = (float *) sigma[j];
        S0ptr = (float *) S0[j];
        Kptr = (float *) K[j];
        answerptr = (float *) answer[j];
        rptr = (float *) r[j];
        cpsptr = (unsigned int *) cps[j];
        for (i=0; i<chunksize_samples; ++i) {
          Tdex   = (i >> 1) & 0x3;
          sigdex = (i >> 3) & 0x3;
          rdex   = (i >> 5) & 0x3;
          S0Kdex = (i >> 7) & 0xf;
          *cpsptr++ = (i&1) ? 0xffffffff : 0;
          *S0ptr++ = S0_array[S0Kdex >> 2];
          *Kptr++ = K_array[S0Kdex];
          *rptr++ = r_array[rdex];
          *sigptr++ = sigma_array[sigdex];
          *Tptr++ = T_array[Tdex];
          *answerptr++ = 0.0;
        }
      }
    }
  }


  nloops = NP >> ((11-double_flag) + LN2_N_SPUS); /* nloops = chunksize_bytes / CHUNK_SIZE */
  
  /* Init ALF Runtime */

  config_parms = NULL;
#ifdef _ALF_PLATFORM_HYBRID_
  sprintf (library_name, "BlackScholes_hybrid_spu64.so");
#else
  if (sizeof(void *) == 4)
    sprintf(library_name, "BlackScholes_cell_spu.so");
  else
    sprintf(library_name, "BlackScholes_cell_spu64.so");
#endif /* _ALF_PLATFORM_HYBRID_ */

  rc = alf_init(config_parms, &alf_handle);
  if(rc < 0){
    fprintf(stderr, "Failed to intialize ALF library, errno: %d\n", rc);
    return -1;
  }

  if (alf_query_system_info(alf_handle,ALF_QUERY_NUM_ACCEL, ALF_ACCEL_TYPE_SPE, &count_spes) < 0){
    fprintf(stderr, "Failed to query available SPEs.\n");
    return -1;
  }
  if (count_spes < (unsigned int)N_SPUS) {
    fprintf(stderr, "System doesn't have enough working SPEs.  I'm leaving.\n");
    return -1;
  }

  if (alf_num_instances_set(alf_handle, N_SPUS) < 0)
  {
    printf ("error in getting spus\n");
    exit (1);

  }
  gettimeofday(&timev1, NULL);

  rc = alf_task_desc_create(alf_handle, 0, &task_desc_handle);
  if(rc < 0){
    fprintf(stderr, "Failed to intialize ALF library, errno: %d\n", rc);
    goto exit;
  }
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(BSPara));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, 6 * CHUNK_SIZE); /* Read cps, S0, K, r, sigma, T */
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, CHUNK_SIZE); /* Only write back answer */
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 8);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 2);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 1);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 4096);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long long)spu_image_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long long)library_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long long)kernel_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long long)input_dtl_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, (unsigned long long)output_dtl_name);
  
  rc =alf_task_create(task_desc_handle, NULL, N_SPUS, ALF_TASK_ATTR_WB_CYCLIC | ALF_TASK_ATTR_SCHED_FIXED, 1, &task_handle);
  if(rc < 0){
    fprintf(stderr, "alf_task_create  error, errno: %d\n", rc);
    goto exit;
  }
#ifdef _ALF_PLATFORM_HYBRID_
 
  /* dataset */
  rc = alf_dataset_create (alf_handle, &dataset_handle);
  if (rc < 0)
  {
    fprintf (stderr, "Call alf_dataset_create error: %d\n", rc);
    goto exit;
  }

  rc = alf_dataset_buffer_add (dataset_handle, cps[0] , 6 * N_SPUS * chunksize_bytes, ALF_DATASET_READ_ONLY);
  if (rc < 0)
  {
    fprintf (stderr, "Call alf_dataset_buffer_add error, cps: %d\n", rc);
    goto exit;
  }
 
  rc = alf_dataset_buffer_add (dataset_handle, answer[0], N_SPUS * chunksize_bytes, ALF_DATASET_WRITE_ONLY);
  if (rc < 0)
  {
    fprintf (stderr, "Call alf_dataset_buffer_add error, answer: %d\n", rc);
    goto exit;
  }

  rc = alf_task_dataset_associate (task_handle, dataset_handle);
  if (rc < 0)
  {
    fprintf (stderr, "Call alf_task_dataset_buffer_associate error, %d\n", rc);
    goto exit;
  }
#endif /* _ALF_PLATFORM_HYBRID_ */

  printf("SPUs started...\n"); 
  
  BSPara.double_flag = (unsigned int)double_flag;
  multi_use_wb_count = nloops;

  /* create N_SPUS number of work blocks, essentially, one work block per SPU */
  for (i = 0; i < N_SPUS; i ++){
    /* Each accelerator instance will process 1 workblock */
    PTR_TO_ADDR64(cps[i], BSPara.cps);
    PTR_TO_ADDR64(S0[i], BSPara.S0);
    PTR_TO_ADDR64(K[i], BSPara.K);
    PTR_TO_ADDR64(r[i], BSPara.r);
    PTR_TO_ADDR64(sigma[i], BSPara.sigma);
    PTR_TO_ADDR64(T[i], BSPara.T);
    PTR_TO_ADDR64(answer[i], BSPara.answer);
    BSPara.chunksize = chunksize_bytes;

    alf_wb_create(task_handle, ALF_WB_MULTI, multi_use_wb_count, &wb_handle);
    alf_wb_parm_add(wb_handle, (void *)&(BSPara.double_flag), 2, ALF_DATA_INT32, 2); /* double_flag & chunksize */
    alf_wb_parm_add(wb_handle, (void *)&(BSPara.cps), 7, ALF_DATA_INT64, 2); /* cps, S0, K, r, sigma, T, answer */
    alf_wb_enqueue(wb_handle);
  }
  rc = alf_task_finalize(task_handle);
  if (rc < 0) {
    fprintf (stderr, "Call alf_task_finalize error, %d\n", rc);
    goto exit;
  }
  
  rc = alf_task_wait(task_handle, -1);
  if (rc < 0) {
    fprintf (stderr, "Call alf_task_wait error, %d\n", rc);
    goto exit;
  }
 
  gettimeofday(&timev2, NULL);

  time_seconds = timev2.tv_sec - timev1.tv_sec;
  time_seconds += 0.000001 * (timev2.tv_usec - timev1.tv_usec);

  printf("Seconds = %f, BlackScholes/Second = %f\n", time_seconds, NP*n_cycles/time_seconds);


  
  printf("checking results...\n"); fflush(stdout);

  hi_error = -1000.0;
  lo_error =  1000.0;
  if (double_flag) {
    double *Tptr, *sigptr, *rptr, *S0ptr, *Kptr, *answerptr;
    unsigned long long *cpsptr;
    for (j=0; j<N_SPUS; ++j) {
      Tptr = (double *) T[j];
      sigptr = (double *) sigma[j];
      S0ptr = (double *) S0[j];
      Kptr = (double *) K[j];
      answerptr = (double *) answer[j];
      rptr = (double *) r[j];
      cpsptr = (unsigned long long *) cps[j];
      for (i=0; i<chunksize_samples; ++i) {
        double d_ans;
        double error;
        d_ans = bspf(*cpsptr++, *S0ptr++, *Kptr++, *rptr++, *sigptr++, *Tptr++);
        error = (*answerptr++) - d_ans;
        if (error > hi_error) hi_error = error;
        if (error < lo_error) {
           lo_error = error;
        }
      }
    }
  }
  else {
    float *Tptr, *sigptr, *rptr, *S0ptr, *Kptr, *answerptr;
    unsigned int *cpsptr;
    for (j=0; j<N_SPUS; ++j) {
      Tptr = (float *) T[j];
      sigptr = (float *) sigma[j];
      S0ptr = (float *) S0[j];
      Kptr = (float *) K[j];
      answerptr = (float *) answer[j];
      rptr = (float *) r[j];
      cpsptr = (unsigned int *) cps[j];
      for (i=0; i<chunksize_samples; ++i) {
        double d_ans;
        double error;
        d_ans = bspf((unsigned long long) *cpsptr++, (double) *S0ptr++, (double) *Kptr++, (double) *rptr++, (double) *sigptr++, (double) *Tptr++);
        error = ((double) *answerptr++) - d_ans;
        if (error > hi_error) hi_error = error;
        if (error < lo_error) {
           lo_error = error;
        }
      }
    }
  }
  if (double_flag) {
     if (lo_error < -7.0e-14 || hi_error > 6.0e-14) {
        printf("error range = %le %le\n", lo_error, hi_error);
        goto exit;
     }
  }
  else {
     if (lo_error < -3.0e-5 || hi_error > 3.0e-5) {
        printf("error range = %le %le\n", lo_error, hi_error);
        goto exit;
     }
  }

  /* tell SPUs to start <ncycles> timing runs */

  printf("performing timing runs...\n"); fflush(stdout);
  gettimeofday(&timev1, NULL);
  
  rc = alf_task_create(task_desc_handle, NULL, N_SPUS, ALF_TASK_ATTR_WB_CYCLIC | ALF_TASK_ATTR_SCHED_FIXED, 1, &task_handle);
  if(rc < 0){
    printf("alf_task_create error: %d\n", rc);
    goto exit;
  }
#ifdef _ALF_PLATFORM_HYBRID_
  rc = alf_task_dataset_associate (task_handle, dataset_handle);
  if (rc < 0) {
    fprintf (stderr, "Call alf_task_dataset_buffer_associate error, %d\n", rc);
    goto exit;
  } 
#endif
  for (n = 0; n < n_cycles; n ++){
    for (i = 0; i < N_SPUS; i ++){
      PTR_TO_ADDR64(cps[i], BSPara.cps);
      PTR_TO_ADDR64(S0[i], BSPara.S0);
      PTR_TO_ADDR64(K[i], BSPara.K);
      PTR_TO_ADDR64(r[i], BSPara.r);
      PTR_TO_ADDR64(sigma[i], BSPara.sigma);
      PTR_TO_ADDR64(T[i], BSPara.T);
      PTR_TO_ADDR64(answer[i], BSPara.answer);
      BSPara.chunksize = chunksize_bytes;

      alf_wb_create(task_handle, ALF_WB_MULTI, multi_use_wb_count, &wb_handle);
      alf_wb_parm_add(wb_handle, (void *)&(BSPara.double_flag), 2, ALF_DATA_INT32, 2); /* double_flag & chunksize */
      alf_wb_parm_add(wb_handle, (void *)&(BSPara.cps), 7, ALF_DATA_INT64, 2); /* cps, S0, K, r, sigma, T, answer */
      alf_wb_enqueue(wb_handle);
    }
  }
  rc = alf_task_finalize(task_handle);
  if (rc < 0) {
    fprintf (stderr, "Call alf_task_finalize error, %d\n", rc);
    goto exit;
  }
  /* Wait and then create task for next cycle */
  alf_task_wait(task_handle, -1);
  if (rc < 0) {
    fprintf (stderr, "Call alf_task_wait error, %d\n", rc);
    goto exit;
  }
  
  gettimeofday(&timev2, NULL);
exit:
  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);


  time_seconds = timev2.tv_sec - timev1.tv_sec;
  time_seconds += 0.000001 * (timev2.tv_usec - timev1.tv_usec);

  printf("Seconds = %f, BlackScholes/Second = %f\n", time_seconds, NP*n_cycles/time_seconds);

  printf("checking results...\n"); fflush(stdout);

  hi_error = -1000.0;
  lo_error =  1000.0;
  if (double_flag) {
    double *Tptr, *sigptr, *rptr, *S0ptr, *Kptr, *answerptr;
    unsigned long long *cpsptr;
    for (j=0; j<N_SPUS; ++j) {
      Tptr = (double *) T[j];
      sigptr = (double *) sigma[j];
      S0ptr = (double *) S0[j];
      Kptr = (double *) K[j];
      answerptr = (double *) answer[j];
      rptr = (double *) r[j];
      cpsptr = (unsigned long long *) cps[j];
      for (i=0; i<chunksize_samples; ++i) {
        double d_ans;
        double error;
        d_ans = bspf(*cpsptr++, *S0ptr++, *Kptr++, *rptr++, *sigptr++, *Tptr++);
        error = (*answerptr++) - d_ans;
        if (error > hi_error) hi_error = error;
        if (error < lo_error) {
           lo_error = error;
        }
      }
    }
  }
  else {
    float *Tptr, *sigptr, *rptr, *S0ptr, *Kptr, *answerptr;
    unsigned int *cpsptr;
    for (j=0; j<N_SPUS; ++j) {
      Tptr = (float *) T[j];
      sigptr = (float *) sigma[j];
      S0ptr = (float *) S0[j];
      Kptr = (float *) K[j];
      answerptr = (float *) answer[j];
      rptr = (float *) r[j];
      cpsptr = (unsigned int *) cps[j];
      for (i=0; i<chunksize_samples; ++i) {
        double d_ans;
        double error;
        d_ans = bspf((unsigned long long) *cpsptr++, (double) *S0ptr++, (double) *Kptr++, (double) *rptr++, (double) *sigptr++, (double) *Tptr++);
        error = ((double) *answerptr++) - d_ans;
        if (error > hi_error) hi_error = error;
        if (error < lo_error) {
           lo_error = error;
        }
      }
    }
  }
  if (double_flag) {
     if (lo_error < -7.0e-14 || hi_error > 6.0e-14) {
        printf("error range = %le %le\n", lo_error, hi_error);
        return 1;
     }
  }
  else {
     if (lo_error < -3.0e-5 || hi_error > 3.0e-5) {
        printf("error range = %le %le\n", lo_error, hi_error);
        return 1;
     }
  }

  remove(mem_file);

  return 0;
}
