/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/* ---------------------------------------------------------------- */
/* Compute PI by Buffon-Needle method based on ALF                  */
/*                                                                  */
/* ---------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <memory.h>
#include <time.h>
#include <byteswap.h>
#include <string.h>

#include "alf.h"
#include "pi.h"

/* Temporary include header file here ends */

#define ERR_THRESHOLD     0.001

result_t result;

#define PATH_BUF_SIZE 1024  // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE];   // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE];     // Used to hold the name of spu library
char spu_image_name[] = "alf_pi_spu";
char kernel_name[] = "comp_kernel";
char ctx_setup_name[] = "context_setup";
char ctx_merge_name[] = "context_merge";

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[PATH_BUF_SIZE];    // Used to hold the complete path to PPU image
#endif

int main(int argc __attribute__ ((__unused__)), char *argv[] __attribute__ ((__unused__)))
{
  /* Declare variables for ALF runtime */
  void *config_parms = NULL;

  alf_handle_t alf_handle;
  alf_task_desc_handle_t desc_info_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;

  udata_t my_udata;
  unsigned int nodes;
  int rc;
  int i;
  double pi;

  result.value = 0.0;
  result.wbs = 0;
  srand(20080808);
  

  // Generate the full path of the SPU image by environment variable
  // ALF_SPU_LIB_PATH
#ifdef _ALF_PLATFORM_HYBRID_
  sprintf(library_name, "alf_pi_hybrid_spu64.so");
#else
  if (sizeof(void *) == 4)
    sprintf(library_name, "alf_pi_cell_spu.so");
  else
    sprintf(library_name, "alf_pi_cell_spu64.so");
#endif

  if (alf_init(config_parms, &alf_handle) < 0) {
    printf("Failed to call alf_init\n");
    return 1;
  }

  if (alf_query_system_info(alf_handle,ALF_QUERY_NUM_ACCEL, ALF_ACCEL_TYPE_SPE, &nodes) < 0)
  {
     printf("Failed to call alf_query_system_info.\n");
     return 1;
  } 
  else if( nodes <= 0 )
  {
     printf("Cannot allocate spe to use.\n");
     return 1;
  }

  rc = alf_num_instances_set(alf_handle, nodes);
  if( !rc ){
    printf("Failed to set ALF instances, instance number is zero!\n");
    goto exit0;
  }

  /* Create a task descriptor for the task */
  rc = alf_task_desc_create(alf_handle, ALF_ACCEL_TYPE_SPE, &desc_info_handle);
  if( rc < 0 ){
     printf("Failed to call alf_task_desc_create, errno: %d\n", rc);
     goto exit0;
  }

  alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(udata_t));
  alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_TSK_CTX_SIZE, sizeof(result_t));
  alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_NUM_DTL, 0);
  alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 0);

  alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long long)spu_image_name);
  alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long long)library_name);
  alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long long)kernel_name);
  alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_CTX_SETUP_REF_L, (unsigned long long)ctx_setup_name);
  alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_CTX_MERGE_REF_L, (unsigned long long)ctx_merge_name);

  alf_task_desc_ctx_entry_add(desc_info_handle, ALF_DATA_DOUBLE, 1);	/* result_t.value */
  alf_task_desc_ctx_entry_add(desc_info_handle, ALF_DATA_INT32, 2);	/* result_t.wbs and result_t.pad */

  // Create task
  rc = alf_task_create(desc_info_handle, &result, 0, 0, 1, &task_handle);
  if(rc <0){
    printf("Failed to call alf_task_create, errno: %d\n", rc);
    goto exit1;
  }

  result.value = 0.0;
  my_udata.iter = ITERATIONS;

  for (i = 0; i < WBS; i++) {
    rc = alf_wb_create(task_handle, ALF_WB_SINGLE, 1, &wb_handle); 
    if(rc <0){
      printf("Failed to call alf_wb_create, errno: %d\n", rc);
      goto exit1;
    }
    my_udata.seed = rand();

    rc = alf_wb_parm_add(wb_handle, (void *)&(my_udata.seed), 1, ALF_DATA_INT32, 2);
    if(rc < 0){
      printf("Failed to call alf_wb_parm_add, errno: %d\n", rc);
      goto exit1;
    }
    rc = alf_wb_parm_add(wb_handle, (void *)&(my_udata.iter), 1, ALF_DATA_INT32, 2);
    if(rc < 0){
      printf("Failed to call alf_wb_parm_add, errno: %d\n", rc);
      goto exit1;
    }
 
    rc = alf_wb_enqueue(wb_handle);
    if(rc < 0){
      printf("Failed to call alf_wb_enqueue, errno: %d\n", rc);
      goto exit1;
    }
  }

  rc = alf_task_finalize(task_handle);
  if(rc < 0){
    printf("Failed to call alf_wb_create, errno: %d\n", rc);
    goto exit1;
  }
 
  /* Wait until the task is finished */
  rc = alf_task_wait(task_handle, -1);
  if(rc < 0){
    printf("Failed to call alf_wb_create, errno: %d\n", rc);
    goto exit1;
  }
  /* Cleanup */
exit1:
  alf_task_desc_destroy(desc_info_handle);
exit0:
  alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);

  pi = result.value / result.wbs;
  printf("PI = %f\n", pi);

  /* Check results */
  if (fabs(pi - 3.14159) >= ERR_THRESHOLD) {
    printf("Get a wrong value of pi\n");
    return -1;
  }

  return rc;
}
