/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/***************************************************************
*   Control  side code
***************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <alf.h>

#include "filter.h"

/*
  * common definition section for a typical ALF application for SDK 3.0 API
  */
const char *app_name = "Iterative Filter";
#define ALF_APP_ID "filter"

#ifdef _ALF_PLATFORM_HYBRID_
    const char *app_library_name_1 = "alf_" ALF_APP_ID "_hybrid_spu64.so";
#elif defined(_ALF_PLATFORM_CELL_)
    #ifdef _ALF_64B_
        const char *app_library_name_1 = "alf_" ALF_APP_ID "_cell_spu64.so";
    #else
        const char *app_library_name_1 = "alf_" ALF_APP_ID "_cell_spu.so";
    #endif
#else
   #error "undefined platform"
#endif

const char *app_image_name_1 = "alf_" ALF_APP_ID "_spu";
const char *app_kernel_name_1 = "comp_kernel";
const char *app_ctx_merge_name_1 = NULL;
const char *app_ctx_setup_name_1 = NULL;
const char *app_prep_in_name_1 = NULL;
const char *app_prep_out_name_1 = NULL;



/*
  * a small init routine to offload the hassle
  */

int app_init_alf(unsigned int num_accel, alf_handle_t *half)
{
    int rc;
    unsigned int nodes;
    void * config_parms = NULL;


      *half = ALF_NULL_HANDLE;
  
      rc = alf_init(config_parms, half);
      if (rc < 0) 
      {
          fprintf(stderr, "**ERR: alf_init failed (%d)\n", rc);
          goto init_err;
      } 
    
      rc = alf_query_system_info(*half,ALF_QUERY_NUM_ACCEL,0, &nodes);
      if (rc < 0) 
      {
          fprintf(stderr, "**ERR: alf_query_system_info failed (%d)\n", rc);
          goto init_err;
      } 
      else if (nodes == 0 || nodes < num_accel) 
      {
          fprintf(stderr, "**ERR: do not have enough accelerators. (have %d, req %d)\n", nodes, num_accel);
          goto init_err;
      }
      
      if(num_accel == 0)
         num_accel = nodes;
      
      rc = alf_num_instances_set(*half, num_accel);
      if(rc < 0)
      {
          fprintf(stderr, "**ERR: alf_num_instances_set(%d) failed (%d)\n", num_accel, rc);
          goto init_err;
      }
      
      return 0;

init_err:

    if(*half)
        alf_exit(*half, ALF_EXIT_POLICY_FORCE, 0);

    *half = ALF_NULL_HANDLE;
    
    return -1;
}




/*
  *  app code begin here 
  */


/* we padded the buffer with additional elements at the begin and end to simplify */
/* the boundary condition check when the calculation comes to a[0] and a[NUM_DATA-1] */

float padded_buffer[2][NUM_DATA+8] __attribute__ ((aligned (16)));

/* the actual buffer pointer used for computation */
float *a[2] = {&padded_buffer[0][4], &padded_buffer[1][4]};  

my_task_context_t tsk_ctx __attribute__ ((aligned(128)));        // task context





/* This function performs data preparation for the application */
/* This function performs data preparation for a[2][NUM_DATA+4] */
void prepare_data()
{
    int i;

	/* init the data with random numbers */
    for(i=0; i<NUM_DATA; i++) {
        a[0][i] = (float)(rand()%8192 - 4096);
    }
    /* initialize to zero the first 4 padding elements */
    for(i=-4; i<0; i++) {
        a[0][i] = (float)0.0f;
        a[1][i] = a[0][i];
    }
    /* initialize to zero the last 4 padding elements */
    for(i=NUM_DATA; i<NUM_DATA+4; i++) {
        a[0][i] = (float)0.0f;
        a[1][i] = a[0][i];
    }
}


int verify_results()
{
    return 0;
}



int main(void)
{
  // basic variables
  alf_handle_t alf_handle;
  int rc;

  // app specific variables 
  alf_task_desc_handle_t task_desc_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;
  alf_wb_sync_handle_t sync_handle;
  my_wb_parms_t  wb_parm;
  unsigned int i,t;
  unsigned int t0, t1; // index of the flip flop buffer


  printf("%s is initializing ALF runtime.\n", app_name); fflush(stdout);
  // basic routines  
  if(app_init_alf(0, &alf_handle) < 0)
  {
      fprintf(stderr, "**ERR: failed to init ALF runtime\n");
      return -1;
  }


  // now comes to the task description and task creation
  rc = alf_task_desc_create(alf_handle, 0, &task_desc_handle);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: alf_task_desc_create (%d)\n", rc);
      exit(-1);
  }

  
  // I'm too tired to check return code of all of these :-)
  // these are common stuffs for most tasks
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long)app_library_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long)app_image_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long)app_kernel_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_CTX_SETUP_REF_L, (unsigned long)app_ctx_setup_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_CTX_MERGE_REF_L, (unsigned long)app_ctx_merge_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long)app_prep_in_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, (unsigned long)app_prep_out_name_1);

  // these are app specific
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, sizeof(my_task_context_t));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(my_wb_parms_t));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, (PART_SIZE+8)*sizeof(float));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, PART_SIZE*sizeof(float));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 16);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 1024);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 0);

  // the task context  
  alf_task_desc_ctx_entry_add(task_desc_handle, ALF_DATA_FLOAT, sizeof(my_task_context_t)/sizeof(float));

  
  printf("%s is initializing input.\n", app_name);fflush(stdout);

  /* prepare data for input */
  prepare_data();

  tsk_ctx.c0 = C0;
  tsk_ctx.c1 = C1;
  tsk_ctx.c2 = C2;

  printf("%s is initialized. Computing results...\n", app_name);

  rc = alf_task_create(task_desc_handle, &tsk_ctx, 0, 0, 0, &task_handle);
  if (rc < 0) 
  {
      fprintf(stderr, "**ERR: cannot create ALF task(%d).\n", rc);
      alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
      return -1;
  }
    
  wb_parm.size = PART_SIZE;

  for(t=0; t<NUM_LOOPS; t++)
  {
	  /* the flip, flop index; */
	  t0 = t&1;
	  t1 = t0^1;
      /* add a barrier before the next iteration to avoid data hazard issues */
	  if(t>0)
		  alf_wb_sync( task_handle, ALF_SYNC_BARRIER, NULL, NULL, 0, &sync_handle );
	  /* creating wb and adding param & io buffer */
	  for (i = 0; i < NUM_DATA; i += PART_SIZE) 
	  {
		alf_wb_create(task_handle, ALF_WB_SINGLE, 0, &wb_handle);
		alf_wb_dtl_begin(wb_handle, ALF_BUF_IN, 0);
		alf_wb_dtl_entry_add(wb_handle, a[t0]+i-4, PART_SIZE, ALF_DATA_FLOAT);
		alf_wb_dtl_end(wb_handle);
		alf_wb_dtl_begin(wb_handle, ALF_BUF_OUT, 0);
		alf_wb_dtl_entry_add(wb_handle, a[t1]+i, PART_SIZE, ALF_DATA_FLOAT);
		alf_wb_dtl_end(wb_handle);
		alf_wb_parm_add(wb_handle, (void *)&wb_parm, sizeof(wb_parm)/sizeof(unsigned int), ALF_DATA_INT32, 0);
		alf_wb_enqueue(wb_handle);
	  }
  }

  alf_task_finalize(task_handle);

  /* alf task wait until wb processed */
  rc = alf_task_wait(task_handle, -1);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: task wait failure(%d).\n", rc);
      alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
      return -1;
  }

  printf("%s is finished. Now verifying results...\n", app_name);

  rc = verify_results();
  if (rc != 0) {
    fprintf(stderr, "**ERR: verification failure(%d).\n", rc);
  } else {
    printf("%s verified successfully\n", app_name);
  }
  
  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);

  return rc;
}



