/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/***************************************************************
*   Control  side code
***************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <alf.h>

#include "dot_prod_multi.h"

/*
  * common definition section for a typical ALF application for SDK 3.0 API
  */

const char *app_name = "Dot Product Using MultiUse WB";
#define ALF_APP_ID "dot_prod_multi"

#ifdef _ALF_PLATFORM_HYBRID_
    const char *app_library_name_1 = "alf_" ALF_APP_ID "_hybrid_spu64.so";
#elif defined(_ALF_PLATFORM_CELL_)
    #ifdef _ALF_64B_
        const char *app_library_name_1 = "alf_" ALF_APP_ID "_cell_spu64.so";
    #else
        const char *app_library_name_1 = "alf_" ALF_APP_ID "_cell_spu.so";
    #endif
#else
   #error "undefined platform"
#endif

const char *app_image_name_1 = "alf_" ALF_APP_ID "_spu";

const char *app_kernel_name_1 = "comp_kernel";
const char *app_ctx_merge_name_1 = NULL;
const char *app_ctx_setup_name_1 = NULL;
const char *app_prep_in_name_1 = "prep_in_dtl";
const char *app_prep_out_name_1 = "prep_out_dtl";


/*
  * a small init routine to offload the hassle
  */

int app_init_alf(unsigned int num_accel, alf_handle_t *half)
{
    int rc;
    unsigned int nodes;
    void * config_parms = NULL;
    
#ifdef _ALF_PLATFORM_HYBRID_
    /* For hybrid, we need to set 1 ppe because of dataset limitations */
    alf_sys_config_t_Hybrid hybrid_config;
    memset(&hybrid_config, 0, sizeof(alf_sys_config_t_Hybrid));
    hybrid_config.num_of_ppes = 1;
    config_parms = (void*)&hybrid_config;
#endif

      *half = ALF_NULL_HANDLE;
  
      rc = alf_init(config_parms, half);
      if (rc < 0) 
      {
          fprintf(stderr, "**ERR: alf_init failed (%d)\n", rc);
          goto init_err;
      } 
    
      rc = alf_query_system_info(*half,ALF_QUERY_NUM_ACCEL,0, &nodes);
      if (rc < 0) 
      {
          fprintf(stderr, "**ERR: alf_query_system_info failed (%d)\n", rc);
          goto init_err;
      } 
      else if (nodes == 0 || nodes < num_accel) 
      {
          fprintf(stderr, "**ERR: do not have enough accelerators. (have %d, req %d)\n", nodes, num_accel);
          goto init_err;
      }
      
      if(num_accel == 0)
         num_accel = nodes;
      
      rc = alf_num_instances_set(*half, num_accel);
      if(rc < 0)
      {
          fprintf(stderr, "**ERR: alf_num_instances_set(%d) failed (%d)\n", num_accel, rc);
          goto init_err;
      }
      
      return 0;

init_err:

    if(*half)
        alf_exit(*half, ALF_EXIT_POLICY_FORCE, 0);

    *half = ALF_NULL_HANDLE;
    
    return -1;
}




/*
  *  app code begin here 
  */


double va[NUM_DATA][VEC_SIZE]  __attribute__ ((aligned (128)));  // the vector a
double vb[NUM_DATA][VEC_SIZE]  __attribute__ ((aligned (128)));  // the vector b
double dot_value[NUM_DATA][2]  __attribute__ ((aligned (128)));  // the return value, padded to 16 byte boundary

                                                       
/* This function performs data preparation for the application */
void prepare_data()
{
    unsigned int i,j;
    
    for(i=0; i<NUM_DATA; i++) 
        for(j=0; j<VEC_SIZE; j++)
        {
            va[i][j] = (double)((rand()&65535) - 32768)/32768.0;
            vb[i][j] = (double)((rand()&65535) - 32768)/32768.0;
        }
}

int verify_results()
{
    unsigned int i,j;
    for(i=0; i<NUM_DATA; i++) 
    {
        double val = 0.0;
        for(j=0; j<VEC_SIZE; j++)
        {
            val += va[i][j]*vb[i][j];
        }
        if(fabs(val - dot_value[i][0]) > fabs(val/10000.0))
        {
            fprintf(stderr, "**ERR: Result verification failed !\n");
            return -1;
        }
    }
    return 0;
}
    




int main(void)
{
  // basic variables
  alf_handle_t alf_handle;
  int rc;

  // app specific variables 
  alf_task_desc_handle_t task_desc_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;
  my_wb_parms_t  wb_parm;
  unsigned int i;


  printf("%s is initializing ALF runtime.\n", app_name); fflush(stdout);
  // basic routines  
  if(app_init_alf(0, &alf_handle) < 0)
  {
      fprintf(stderr, "**ERR: failed to init ALF runtime\n");
      return -1;
  }


  // now comes to the task description and task creation
  rc = alf_task_desc_create(alf_handle, 0, &task_desc_handle);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: alf_task_desc_create (%d)\n", rc);
      exit(-1);
  }

  
  // I'm too tired to check return code of all of these :-)
  // these are common stuffs for most tasks
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long)app_library_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long)app_image_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long)app_kernel_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_CTX_SETUP_REF_L, (unsigned long)app_ctx_setup_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_CTX_MERGE_REF_L, (unsigned long)app_ctx_merge_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long)app_prep_in_name_1);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, (unsigned long)app_prep_out_name_1);

  // these are app specific
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(my_wb_parms_t));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, PART_SIZE*sizeof(double)*2);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, 16);  // 16 bytes
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 16);  // this should be enough in our specific case
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 2);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 1024);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 1); // DTL gen on Accel

  
  printf("%s is initializing input.\n", app_name);fflush(stdout);

  /* prepare data for input */
  prepare_data();

  printf("%s is initialized. Computing results...\n", app_name);
  
  rc = alf_task_create(task_desc_handle, NULL, 0, 0,  0, &task_handle);
  if (rc < 0) 
  {
      fprintf(stderr, "**ERR: cannot create ALF task(%d).\n", rc);
      alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
      return -1;
  }
  
  alf_dataset_handle_t dataset_handle;
  alf_dataset_create(alf_handle, &dataset_handle);
  alf_dataset_buffer_add(dataset_handle, va, NUM_DATA*VEC_SIZE*sizeof(double), ALF_DATASET_READ_ONLY);
  alf_dataset_buffer_add(dataset_handle, vb, NUM_DATA*VEC_SIZE*sizeof(double), ALF_DATASET_READ_ONLY);
  alf_dataset_buffer_add(dataset_handle, dot_value, NUM_DATA*2*sizeof(double), ALF_DATASET_READ_WRITE);
  alf_task_dataset_associate(task_handle, dataset_handle);

  /* creating wb and adding param & io buffer */
  for (i = 0; i < NUM_DATA; i++)
  {
      wb_parm.accumulator = 0.0;
      wb_parm.num_data = VEC_SIZE;
      wb_parm.part_size = PART_SIZE;
#ifdef _ALF_64B_
      wb_parm.va_addr = (alf_data_addr64_t)&va[i][0];
      wb_parm.vb_addr = (alf_data_addr64_t)&vb[i][0];
      wb_parm.c_addr = (alf_data_addr64_t)&dot_value[i][0];
#else  // 32 bit address
      wb_parm.va_addr = (alf_data_addr64_t)(alf_data_addr32_t)&va[i][0];
      wb_parm.vb_addr = (alf_data_addr64_t)(alf_data_addr32_t)&vb[i][0];
      wb_parm.c_addr = (alf_data_addr64_t)(alf_data_addr32_t)&dot_value[i][0];
#endif
      /* we use multi use work block for the full vector */
	  alf_wb_create(task_handle, ALF_WB_MULTI,
		  (VEC_SIZE+PART_SIZE-1)/PART_SIZE, /* the repeat count */
		  &wb_handle);
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.num_data, 2, ALF_DATA_INT32, 0); // num_data and part_size
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.va_addr, 3, ALF_DATA_ADDR64, 0); // va_addr, vb_addr, c_addr
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.accumulator, 1, ALF_DATA_DOUBLE, 0);  // accumulator
      alf_wb_enqueue(wb_handle);
  }
  
  alf_task_finalize(task_handle);

  /* alf task wait until wb processed */
  rc = alf_task_wait(task_handle, -1);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: task wait failure(%d).\n", rc);
      alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
      return -1;
  }

  printf("%s is finished. Now verifying results...\n", app_name);

  rc = verify_results();
  if (rc != 0) {
    fprintf(stderr, "**ERR: verification failure(%d).\n", rc);
  } else {
    printf("%s verified successfully\n", app_name);
  }
  
  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);

  return rc;
}



