/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/***************************************************************
*   Accelerator side code
***************************************************************/

#include <alf_accel.h>
#include "../dot_prod_multi.h"

static unsigned int calc_data_size(my_wb_parms_t *p_parm, unsigned int current_count)
{
    // do not give me wrong data, I do not check for it
    if((current_count+1)*p_parm->part_size <= p_parm->num_data)
        return p_parm->part_size;  // full block
    else
        return p_parm->num_data - p_parm->part_size*current_count;  // remainning partial block
}

/* the computation kernel function */
int comp_kernel(void *p_task_context  __attribute__ ((unused)),
                    void *p_parm_ctx_buffer,
                    void *p_input_buffer,
                    void *p_output_buffer,
                    void *p_inout_buffer  __attribute__ ((unused)),
                    unsigned int current_count,
                    unsigned int total_count)
{
    my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_ctx_buffer;

    double *va, *vb;
    double *out = (double *)p_output_buffer;
    double val=0.0;
    unsigned int size;
    unsigned int i;

    // calculate the actual block size
    size = calc_data_size(p_parm, current_count);
    
    va = (double *)p_input_buffer;
    vb = va + size;
    
    // I won't do any vector optimization here, for readability
    for(i=0;i<size;i++) 
    {
       val += va[i]*vb[i];
    }
    
    p_parm->accumulator += val;  // accumulation into parm_ctx_buffer

    if(current_count+1 == total_count)  // at the last iteration, we have the right result now
    {
        *out = p_parm->accumulator;
    }
 
    return 0;
}

int prep_in_dtl(void *p_task_context __attribute__ ((unused)),
                 void *p_parm_ctx_buffer,
                 void *p_dtl,
                 unsigned int current_count,
                 unsigned int total_count __attribute__ ((unused)))
{
    my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_ctx_buffer;
    unsigned int size, k;

    // calculate the actual block size
    size = calc_data_size(p_parm, current_count);

    ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_IN, 0);
    for(k=0; k<size; k+=2048)  // Cell Specific 16KB issue, 2048*double is 16KB
    {
       unsigned int bsize;
       if(k+2048 <= size)
           bsize = 2048;
       else
           bsize = size - k;
       ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, bsize, ALF_DATA_DOUBLE, p_parm->va_addr + (current_count*p_parm->part_size + k)*sizeof(double));
    }
    for(k=0; k<size; k+=2048)  // Cell Specific 16KB issue, 2048*double is 16KB
    {
       unsigned int bsize;
       if(k+2048 <= size)
           bsize = 2048;
       else
           bsize = size - k;
       ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, bsize, ALF_DATA_DOUBLE, p_parm->vb_addr + (current_count*p_parm->part_size + k)*sizeof(double));
    }
    ALF_ACCEL_DTL_END(p_dtl);

    return 0;
}

int prep_out_dtl(void *p_task_context __attribute__ ((unused)),
                   void *p_parm_ctx_buffer,
                   void *p_dtl,
                   unsigned int current_count,
                   unsigned int total_count )
{
    my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_ctx_buffer;
    if(current_count+1 == total_count)  // at the last iteration, we have the right result now
    {
        ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_OUT, 0);
        ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, 1, ALF_DATA_DOUBLE, p_parm->c_addr);
        ALF_ACCEL_DTL_END(p_dtl);
    }
    return 0;
}




ALF_ACCEL_EXPORT_API_LIST_BEGIN
   ALF_ACCEL_EXPORT_API("comp_kernel", comp_kernel);
   ALF_ACCEL_EXPORT_API("prep_in_dtl", prep_in_dtl);
   ALF_ACCEL_EXPORT_API("prep_out_dtl", prep_out_dtl);
ALF_ACCEL_EXPORT_API_LIST_END


